import toolviper.utils.logger as logger
import xarray as xr


def convert_generic_xds_to_xradio_schema(
    generic_xds: xr.Dataset,
    msv4_xds: xr.Dataset,
    to_new_data_variables: dict[str, list],
    to_new_coords: dict[str, list],
) -> xr.Dataset:
    """Converts a generic xarray Dataset to the xradio schema.

    This function takes a generic xarray Dataset and converts it to an xradio schema
    represented by the msv4_xds Dataset. It performs the conversion based on the provided
    mappings in the to_new_data_variables and to_new_coords dictionaries.

    Parameters
    ----------
    generic_xds : xr.Dataset
        The generic xarray Dataset to be converted.
    msv4_xds : xr.Dataset
        The xradio schema represented by the msv4_xds Dataset.
    to_new_data_variables : dict
        A dictionary mapping the data variables/coordinates in the generic_xds Dataset to the new data variables
        in the msv4_xds Dataset. The keys are the old data variables/coordinates and the values are a list of the new name and a list of the new dimension names.
    to_new_coords : dict
        A dictionary mapping  data variables/coordinates in the generic_xds Dataset to the new coordinates
        in the msv4_xds Dataset. The keys are the old data variables/coordinates and the values are a list of the new name and a list of the new dimension names.

    Returns
    -------
    xr.Dataset
        The converted xradio schema represented by the msv4_xds Dataset.

    Notes
    -----
    Example to_new_data_variables:
    to_new_data_variables = {
        "POSITION": ["ANTENNA_POSITION",["name", "cartesian_pos_label"]],
        "OFFSET": ["ANTENNA_FEED_OFFSET",["name", "cartesian_pos_label"]],
        "DISH_DIAMETER": ["ANTENNA_DISH_DIAMETER",["name"]],
    }

    Example to_new_coords:
    to_new_coords = {
        "NAME": ["name",["name"]],
        "STATION": ["station",["name"]],
        "MOUNT": ["mount",["name"]],
        "PHASED_ARRAY_ID": ["phased_array_id",["name"]],
        "antenna_id": ["antenna_id",["name"]],
    }
    """

    column_description = generic_xds.attrs["other"]["msv2"]["ctds_attrs"][
        "column_descriptions"
    ]
    coords = {}
    coord_attrs = {}

    name_keys = list(generic_xds.data_vars.keys()) + list(generic_xds.coords.keys())

    for key in name_keys:

        if key in column_description:
            msv4_measure = column_description_casacore_to_msv4_measure(
                column_description[key]
            )
        else:
            msv4_measure = None

        if key in to_new_data_variables:
            new_dv = to_new_data_variables[key]
            msv4_xds[new_dv[0]] = xr.DataArray(generic_xds[key].data, dims=new_dv[1])

            if msv4_measure:
                msv4_xds[new_dv[0]].attrs.update(msv4_measure)

        if key in to_new_coords:
            new_coord = to_new_coords[key]
            coords[new_coord[0]] = (
                new_coord[1],
                generic_xds[key].data,
            )

            if msv4_measure:
                coord_attrs[new_coord[0]] = msv4_measure

    msv4_xds = msv4_xds.assign_coords(coords)
    for coord, coord_attrs in coord_attrs.items():
        msv4_xds.coords[coord].attrs.update(coord_attrs)

    return msv4_xds


def column_description_casacore_to_msv4_measure(
    casacore_column_description, ref_code=None, time_format="unix"
):
    import numpy as np

    msv4_measure = {}
    if "MEASINFO" in casacore_column_description["keywords"]:
        measinfo = casacore_column_description["keywords"]["MEASINFO"]

        # Get conversion information
        msv4_measure_conversion = casacore_to_msv4_measure_type[measinfo["type"]]

        # Convert type, copy unit
        msv4_measure["type"] = msv4_measure_conversion["type"]
        msv4_measure["units"] = list(
            casacore_column_description["keywords"]["QuantumUnits"]
        )

        # Beware: casa_ref won't be found in cases such as the custom
        # 'NRAO_GBT_USER/NRAO_GBT_USER_DIR_REF' in POINTING
        casa_ref = None
        # Reference frame to convert?
        if "Ref" in msv4_measure_conversion:
            # Find reference frame
            if "TabRefCodes" in measinfo:
                ref_index = np.where(measinfo["TabRefCodes"] == ref_code)[0][0]
                casa_ref = measinfo["TabRefTypes"][ref_index]
            elif "Ref" in measinfo:
                casa_ref = measinfo["Ref"]
            elif measinfo["type"] == "frequency":
                # Some MSv2 don't have the "TabRefCodes".
                ref_index = np.where(casa_frequency_frames_codes == ref_code)[0][0]
                casa_ref = casa_frequency_frames[ref_index]
            else:
                logger.debug(
                    f"Could not determine {measinfo['type']} measure "
                    "reference frame!"
                )

            # Convert into MSv4 representation of reference frame, warn if unknown
            if casa_ref in msv4_measure_conversion.get("Ref_map", {}):
                casa_ref = msv4_measure_conversion["Ref_map"][casa_ref]
            else:
                logger.debug(
                    f"No translation defined for casacore reference frame ({measinfo['type']} "
                    f"measure), using verbatim: {casa_ref}"
                )
                if measinfo["type"] == "direction" and casa_ref == "B1950_VLA":
                    raise RuntimeError(
                        f"The direction reference frame {casa_ref} is not supported. "
                        "Please use CASA to convert the frame to supported frames like J2000 or ICRS."
                    )

            msv4_measure[msv4_measure_conversion["Ref"]] = casa_ref

        if msv4_measure["type"] == "time":
            msv4_measure["format"] = time_format
    elif "QuantumUnits" in casacore_column_description["keywords"]:
        msv4_measure = {
            "type": "quantity",
            "units": list(casacore_column_description["keywords"]["QuantumUnits"]),
        }

    return msv4_measure


# Map casacore measures to astropy
casacore_to_msv4_measure_type = {
    "quanta": {
        "type": "quantity",
    },
    "direction": {
        "type": "sky_coord",
        "Ref": "frame",
        "Ref_map": {
            "AZELGEO": "altaz",
            "ICRS": "icrs",
            "J2000": "fk5",
        },
    },
    "epoch": {
        "type": "time",
        "Ref": "scale",
        "Ref_map": {
            "UTC": "utc",
            "TAI": "tai",
        },
    },
    "frequency": {
        "type": "spectral_coord",
        "Ref": "observer",
        "Ref_map": {
            "REST": "REST",
            "LSRK": "lsrk",
            "LSRD": "lsrd",
            "BARY": "BARY",
            "GEO": "GEO",
            "TOPO": "TOPO",
        },  # The frames/observer we are not sure if/how to translate to astropy are uppercase
    },
    "position": {
        "type": "location",
        "Ref": "frame",
        "Ref_map": {"ITRF": "GRS80"},
    },
    "uvw": {
        "type": "uvw",
        "Ref": "frame",
        "Ref_map": {"J2000": "fk5", "APP": "APP"},
    },
    "radialvelocity": {"type": "quantity"},
}

casa_frequency_frames = [
    "REST",
    "LSRK",
    "LSRD",
    "BARY",
    "GEO",
    "TOPO",
    "GALACTO",
    "LGROUP",
    "CMB",
    "Undefined",
]

casa_frequency_frames_codes = [0, 1, 2, 3, 4, 5, 6, 7, 8, 64]



import numpy as np

_deg_to_rad = np.pi / 180


def haversine(ra1, dec1, ra2, dec2):
    """
    Calculate the great circle distance between two points
    on the celestial sphere specified in radians.
    """
    # Haversine formula
    d_ra = ra2 - ra1
    d_dec = dec2 - dec1
    a = np.sin(d_dec / 2) ** 2 + np.cos(dec1) * np.cos(dec2) * np.sin(d_ra / 2) ** 2
    c = 2 * np.arcsin(np.sqrt(a))
    return c


def add_position_offsets(dv_1, dv_2):
    # Fun with angles: We are adding angles together. We need to make sure that the results are between -pi and pi.
    new_pos = dv_1 + dv_2

    new_pos = wrap_to_pi(new_pos)

    # while np.any(new_pos[:, 0] > np.pi) or np.any(new_pos[:, 0] < -np.pi):
    #     new_pos[:, 0] = np.where(
    #         new_pos[:, 0] > np.pi, new_pos[:, 0] - 2 * np.pi, new_pos[:, 0]
    #     )
    #     new_pos[:, 0] = np.where(
    #         new_pos[:, 0] < -np.pi, new_pos[:, 0] + 2 * np.pi, new_pos[:, 0]
    #     )

    # while np.any(new_pos[:, 1] > np.pi / 2) or np.any(new_pos[:, 1] < -np.pi / 2):
    #     new_pos[:, 1] = np.where(
    #         new_pos[:, 1] > np.pi / 2, new_pos[:, 1] - np.pi, new_pos[:, 1]
    #     )
    #     new_pos[:, 1] = np.where(
    #         new_pos[:, 1] < -np.pi / 2, new_pos[:, 1] + np.pi, new_pos[:, 1]
    #     )

    return new_pos


def wrap_to_pi(angles):
    """
    Wraps an array of angles in radians to the range (-π, π].

    Parameters:
    -----------
    angles : array_like
        Input array of angles in radians.

    Returns:
    --------
    array_like
        The input angles wrapped to the range (-π, π].

    Examples:
    ---------
    >>> wrap_to_pi(np.array([0, np.pi, -np.pi, 3*np.pi, -3*np.pi]))
    array([ 0.        ,  3.14159265, -3.14159265,  3.14159265, -3.14159265])
    """
    return (angles + np.pi) % (2 * np.pi) - np.pi


def wrap_to_pi(angles):
    return (angles + np.pi) % (2 * np.pi) - np.pi


def convert_to_si_units(xds):
    for data_var in xds.data_vars:
        if "units" in xds[data_var].attrs:
            for u_i, u in enumerate(xds[data_var].attrs["units"]):
                if u == "km":
                    xds[data_var][..., u_i] = xds[data_var][..., u_i] * 1e3
                    xds[data_var].attrs["units"][u_i] = "m"
                if u == "km/s":
                    xds[data_var][..., u_i] = xds[data_var][..., u_i] * 1e3
                    xds[data_var].attrs["units"][u_i] = "m/s"
                if u == "deg":
                    xds[data_var][..., u_i] = xds[data_var][..., u_i] * np.pi / 180
                    xds[data_var].attrs["units"][u_i] = "rad"
                if u == "Au" or u == "AU":
                    xds[data_var][..., u_i] = xds[data_var][..., u_i] * 149597870700
                    xds[data_var].attrs["units"][u_i] = "m"
                if u == "Au/d" or u == "AU/d":
                    xds[data_var][..., u_i] = (
                        xds[data_var][..., u_i] * 149597870700 / 86400
                    )
                    xds[data_var].attrs["units"][u_i] = "m/s"
                if u == "arcsec":
                    xds[data_var][..., u_i] = xds[data_var][..., u_i] * np.pi / 648000
                    xds[data_var].attrs["units"][u_i] = "rad"
                if u == "hPa":
                    xds[data_var][..., u_i] = xds[data_var][..., u_i] * 100.0
                    xds[data_var].attrs["units"][u_i] = "Pa"
                if u == "m-2":
                    # IONOS_ELECTRON sometimes has "m-2" instead of "/m^2"
                    xds[data_var].attrs["units"][u_i] = "/m^2"
    return xds



import astropy as ap
from astropy import units as u
from astropy.io import fits
from astropy.time import Time
from ..common import (
    _compute_linear_world_values,
    _compute_velocity_values,
    _compute_world_sph_dims,
    _convert_beam_to_rad,
    _default_freq_info,
    _doppler_types,
    _freq_from_vel,
    _get_unit,
    _get_xds_dim_order,
    _image_type,
    _l_m_attr_notes,
)
from xradio._utils.coord_math import _deg_to_rad
from xradio._utils.dict_helpers import make_quantity
import copy
import dask
import dask.array as da
import numpy as np
import re
from typing import Union
import xarray as xr


def _fits_image_to_xds(
    img_full_path: str, chunks: dict, verbose: bool, do_sky_coords: bool
) -> dict:
    """
    TODO: complete documentation
    Create an xds without any pixel data from metadata from the specified FITS image
    """
    # memmap = True allows only part of data to be loaded into memory
    # may also need to pass mode='denywrite'
    # https://stackoverflow.com/questions/35759713/astropy-io-fits-read-row-from-large-fits-file-with-mutliple-hdus
    hdulist = fits.open(img_full_path, memmap=True)
    attrs, helpers, header = _fits_header_to_xds_attrs(hdulist)
    hdulist.close()
    # avoid keeping reference to mem-mapped fits file
    del hdulist
    xds = _create_coords(helpers, header, do_sky_coords)
    sphr_dims = helpers["sphr_dims"]
    ary = _read_image_array(img_full_path, chunks, helpers, verbose)
    dim_order = _get_xds_dim_order(sphr_dims)
    xds = _add_sky_or_aperture(xds, ary, dim_order, helpers, sphr_dims)
    xds.attrs = attrs
    xds = _add_coord_attrs(xds, helpers)
    if helpers["has_multibeam"]:
        xds = _do_multibeam(xds, img_full_path)
    return xds


def _add_coord_attrs(xds: xr.Dataset, helpers: dict) -> xr.Dataset:
    xds = _add_time_attrs(xds, helpers)
    xds = _add_freq_attrs(xds, helpers)
    xds = _add_vel_attrs(xds, helpers)
    xds = _add_l_m_attrs(xds, helpers)
    xds = _add_lin_attrs(xds, helpers)
    return xds


def _add_time_attrs(xds: xr.Dataset, helpers: dict) -> xr.Dataset:
    time_coord = xds.coords["time"]
    meta = copy.deepcopy(helpers["obsdate"])
    del meta["value"]
    # meta['format'] = 'MJD'
    # meta['time_scale'] = meta['refer']
    # del meta['refer']
    time_coord.attrs = meta
    xds.assign_coords(time=time_coord)
    return xds


def _add_freq_attrs(xds: xr.Dataset, helpers: dict) -> xr.Dataset:
    freq_coord = xds.coords["frequency"]
    meta = {}
    if helpers["has_freq"]:
        meta["rest_frequency"] = make_quantity(helpers["restfreq"], "Hz")
        meta["frame"] = helpers["specsys"]
        meta["units"] = "Hz"
        meta["type"] = "frequency"
        meta["wave_unit"] = "mm"
        freq_axis = helpers["freq_axis"]
        meta["crval"] = helpers["crval"][freq_axis]
        meta["cdelt"] = helpers["cdelt"][freq_axis]
    if not meta:
        # this is the default frequency information CASA creates
        meta = _default_freq_info()
    freq_coord.attrs = copy.deepcopy(meta)
    xds["frequency"] = freq_coord
    return xds


def _add_vel_attrs(xds: xr.Dataset, helpers: dict) -> xr.Dataset:
    vel_coord = xds.coords["velocity"]
    meta = {"units": "m/s"}
    if helpers["has_freq"]:
        meta["doppler_type"] = helpers.get("doppler", "RADIO")
    else:
        meta["doppler_type"] = _doppler_types[0]
    meta["type"] = "doppler"
    vel_coord.attrs = copy.deepcopy(meta)
    xds.coords["velocity"] = vel_coord
    return xds


def _add_l_m_attrs(xds: xr.Dataset, helpers: dict) -> xr.Dataset:
    attr_note = _l_m_attr_notes()
    for c in ["l", "m"]:
        if c in xds.coords:
            xds[c].attrs = {
                "crval": 0.0,
                "cdelt": helpers[c]["cdelt"],
                "units": "rad",
                "type": "quantity",
                "note": attr_note[c],
            }
    return xds


def _add_lin_attrs(xds: xr.Dataset, helpers: dict) -> xr.Dataset:
    """
    if helpers["sphr_dims"]:
        for i, name in zip(helpers["dir_axes"], helpers["sphr_axis_names"]):
            meta = {
                "units": "rad",
                "crval": helpers["crval"][i],
                "cdelt": helpers["cdelt"][i],
            }
            xds.coords[name].attrs = meta
    """
    if not helpers["sphr_dims"]:
        for i, j in zip(helpers["dir_axes"], ("u", "v")):
            meta = {
                "units": "wavelengths",
                "crval": helpers["crval"][i],
                "cdelt": helpers["cdelt"][i],
            }
            xds.coords[j].attrs = meta
    return xds


def _is_freq_like(v: str) -> bool:
    return v.startswith("FREQ") or v == "VOPT" or v == "VRAD"


def _xds_direction_attrs_from_header(helpers: dict, header) -> dict:
    # helpers is modified in place, headers is not modified
    t_axes = helpers["t_axes"]
    p0 = header[f"CTYPE{t_axes[0]}"][-3:]
    p1 = header[f"CTYPE{t_axes[1]}"][-3:]
    if p0 != p1:
        raise RuntimeError(
            f"Projections for direction axes ({p0}, {p1}) differ, but they "
            "must be the same"
        )
    direction = {}
    direction["projection"] = p0
    helpers["projection"] = p0
    ref_sys = header["RADESYS"]
    ref_eqx = header["EQUINOX"]
    if ref_sys == "FK5" and ref_eqx == 2000:
        ref_eqx = "J2000.0"
    helpers["ref_sys"] = ref_sys
    helpers["ref_eqx"] = ref_eqx
    # fits does not support conversion frames
    direction["reference"] = {
        "type": "sky_coord",
        "frame": ref_sys,
        "equinox": ref_eqx,
        "units": ["rad", "rad"],
        "value": [0.0, 0.0],
    }
    dir_axes = helpers["dir_axes"]
    for i in dir_axes:
        x = helpers["crval"][i] * u.Unit(_get_unit(helpers["cunit"][i]))
        x = x.to("rad")
        direction["reference"]["value"][i] = x.value
        x = helpers["cdelt"][i] * u.Unit(_get_unit(helpers["cunit"][i]))
        x = x.to("rad")
    direction["latpole"] = make_quantity(header["LATPOLE"] * _deg_to_rad, "rad")
    direction["longpole"] = make_quantity(header["LONPOLE"] * _deg_to_rad, "rad")
    pc = np.zeros([2, 2])
    for i in (0, 1):
        for j in (0, 1):
            # dir_axes are now 0-based, but fits needs 1-based
            pc[i][j] = header[f"PC{dir_axes[i]+1}_{dir_axes[j]+1}"]
    direction["pc"] = pc
    # Is there really no fits header parameter for projection_parameters?
    direction["projection_parameters"] = np.array([0.0, 0.0])
    return direction


def _fits_header_c_values_to_metadata(helpers: dict, header) -> None:
    # The helpers dict is modified in place. header is not modified
    ctypes = []
    shape = []
    crval = []
    cdelt = []
    crpix = []
    cunit = []
    for i in range(1, helpers["naxes"] + 1):
        ax_type = header[f"CTYPE{i}"]
        ctypes.append(ax_type)
        shape.append(header[f"NAXIS{i}"])
        crval.append(header[f"CRVAL{i}"])
        cdelt.append(header[f"CDELT{i}"])
        # FITS 1-based to python 0-based
        crpix.append(header[f"CRPIX{i}"] - 1)
        cunit.append(header[f"CUNIT{i}"])
    helpers["ax_type"] = ax_type
    helpers["shape"] = shape
    helpers["ctype"] = ctypes
    helpers["crval"] = crval
    helpers["cdelt"] = cdelt
    helpers["crpix"] = crpix
    helpers["cunit"] = cunit


def _get_telescope_metadata(helpers: dict, header) -> dict:
    # The helpers dict is modified in place. header is not modified
    tel = {}
    tel["name"] = header["TELESCOP"]
    if "OBSGEO-x" in header:
        x = header["OBSGEO-X"]
        y = header["OBSGEO-Y"]
        z = header["OBSGEO-Z"]
        xyz = np.array([x, y, z])
        r = np.sqrt(np.sum(xyz * xyz))
        lat = np.arcsin(z / r)
        long = np.arctan2(y, x)
        tel["position"] = {
            "type": "position",
            # I haven't seen a FITS keyword for reference frame of telescope posiiton
            "ellipsoid": "GRS80",
            "units": ["rad", "rad", "m"],
            "value": np.array([long, lat, r]),
        }
        helpers["tel_pos"] = tel["position"]
    return tel


def _pointing_center_to_metadata(helpers: dict, header) -> dict:
    # Neither helpers or header is modified
    t_axes = helpers["t_axes"]
    long_unit = header[f"CUNIT{t_axes[0]}"]
    lat_unit = header[f"CUNIT{t_axes[1]}"]
    unit = []
    for uu in [long_unit, lat_unit]:
        new_u = u.Unit(_get_unit(uu))
        unit.append(new_u)
    pc_long = float(header[f"CRVAL{t_axes[0]}"]) * unit[0]
    pc_lat = float(header[f"CRVAL{t_axes[1]}"]) * unit[1]
    pc_long = pc_long.to(u.rad).value
    pc_lat = pc_lat.to(u.rad).value
    return {"value": np.array([pc_long, pc_lat]), "initial": True}


def _user_attrs_from_header(header) -> dict:
    # header is not modified
    exclude = [
        "ALTRPIX",
        "ALTRVAL",
        "BITPIX",
        "BSCALE",
        "BTYPE",
        "BUNIT",
        "BZERO",
        "CASAMBM",
        "DATE",
        "DATE-OBS",
        "EQUINOX",
        "EXTEND",
        "HISTORY",
        "LATPOLE",
        "LONPOLE",
        "OBSERVER",
        "ORIGIN",
        "TELESCOP",
        "OBJECT",
        "RADESYS",
        "RESTFRQ",
        "SIMPLE",
        "SPECSYS",
        "TIMESYS",
        "VELREF",
    ]
    regex = r"|".join(
        [
            "^NAXIS\\d?$",
            "^CRVAL\\d$",
            "^CRPIX\\d$",
            "^CTYPE\\d$",
            "^CDELT\\d$",
            "^CUNIT\\d$",
            "^OBSGEO-(X|Y|Z)$",
            "^P(C|V)\\d_\\d$",
        ]
    )
    user = {}
    for k, v in header.items():
        if re.search(regex, k) or k in exclude:
            continue
        user[k.lower()] = v
    return user


def _beam_attr_from_header(helpers: dict, header) -> Union[dict, str, None]:
    # The helpers dict is modified in place. header is not modified
    helpers["has_multibeam"] = False
    if "BMAJ" in header:
        # single global beam
        beam = {
            "bmaj": make_quantity(header["BMAJ"], "arcsec"),
            "bmin": make_quantity(header["BMIN"], "arcsec"),
            "pa": make_quantity(header["BPA"], "arcsec"),
        }
        return _convert_beam_to_rad(beam)
    elif "CASAMBM" in header and header["CASAMBM"]:
        # multi-beam
        helpers["has_multibeam"] = True
        return "mb"
    else:
        # no beam
        return None


def _create_dim_map(helpers: dict, header) -> dict:
    # The helpers dict is modified in place. header is not modified
    t_axes = np.array([0, 0])
    dim_map = {}
    helpers["has_freq"] = False
    # fits indexing starts at 1, not 0
    for i in range(1, helpers["naxes"] + 1):
        ax_type = header[f"CTYPE{i}"]
        if ax_type.startswith("RA-"):
            t_axes[0] = i
        elif ax_type.startswith("DEC-"):
            t_axes[1] = i
        elif ax_type == "STOKES":
            dim_map["polarization"] = i - 1
        elif _is_freq_like(ax_type):
            dim_map["freq"] = i - 1
            helpers["has_freq"] = True
            # helpers["native_type"] = ax_type
        else:
            raise RuntimeError(f"{ax_type} is an unsupported axis")
    helpers["t_axes"] = t_axes
    helpers["dim_map"] = dim_map
    return dim_map


def _fits_header_to_xds_attrs(hdulist: fits.hdu.hdulist.HDUList) -> dict:
    primary = None
    beams = None
    for hdu in hdulist:
        if hdu.name == "PRIMARY":
            primary = hdu
        elif hdu.name == "BEAMS":
            beams = hdu
        else:
            raise RuntimeError(f"Unknown HDU name {hdu.name}")
    if not primary:
        raise RuntimeError(f"No PRIMARY HDU found in fits file")
    header = primary.header
    helpers = {}
    attrs = {}
    naxes = header["NAXIS"]
    helpers["naxes"] = naxes
    dim_map = _create_dim_map(helpers, header)
    _fits_header_c_values_to_metadata(helpers, header)
    if "RESTFRQ" in header:
        helpers["restfreq"] = header["RESTFRQ"]
    if "SPECSYS" in header:
        helpers["specsys"] = header["SPECSYS"]
    t_axes = helpers["t_axes"]
    if (t_axes > 0).all():
        dir_axes = t_axes[:]
        dir_axes = dir_axes - 1
        helpers["dir_axes"] = dir_axes
        dim_map["l"] = dir_axes[0]
        dim_map["m"] = dir_axes[1]
        helpers["dim_map"] = dim_map
    else:
        raise RuntimeError("Could not find both direction axes")
    if dir_axes is not None:
        attrs["direction"] = _xds_direction_attrs_from_header(helpers, header)
    # FIXME read fits data in chunks in case all data too large to hold in memory
    has_mask = da.any(da.isnan(primary.data)).compute()
    attrs["active_mask"] = "mask0" if has_mask else None
    helpers["has_mask"] = has_mask
    beam = _beam_attr_from_header(helpers, header)
    if beam != "mb":
        attrs["beam"] = beam
    if "BITPIX" in header:
        v = abs(header["BITPIX"])
        if v == 32:
            helpers["dtype"] = "float32"
        elif v == 64:
            helpers["dtype"] = "float64"
        else:
            raise RuntimeError(f'Unhandled data type {header["BITPIX"]}')
    helpers["btype"] = header["BTYPE"] if "BTYPE" in header else None
    helpers["bunit"] = header["BUNIT"] if "BUNIT" in header else None
    attrs["object_name"] = header["OBJECT"] if "OBJECT" in header else None
    obsdate = {}
    obsdate["type"] = "time"
    obsdate["value"] = Time(header["DATE-OBS"], format="isot").mjd
    obsdate["units"] = "d"
    obsdate["scale"] = header["TIMESYS"]
    obsdate["format"] = "MJD"
    attrs["obsdate"] = obsdate
    helpers["obsdate"] = obsdate
    attrs["observer"] = header["OBSERVER"]
    attrs["pointing_center"] = _pointing_center_to_metadata(helpers, header)
    attrs["description"] = None
    attrs["telescope"] = _get_telescope_metadata(helpers, header)
    # TODO complete _make_history_xds when spec has been finalized
    # attrs['history'] = _make_history_xds(header)
    attrs["user"] = _user_attrs_from_header(header)
    return attrs, helpers, header


def _make_history_xds(header):
    # TODO complete writing history when we actually have a spec for what
    # the image history is supposed to be, since doing this now may
    # be a waste of time if the final spec turns out to be significantly
    # different from our current ad hoc history xds
    # in astropy, 3506803168 seconds corresponds to 1970-01-01T00:00:00
    history_list = list(header.get("HISTORY"))
    for i in range(len(history_list) - 1, -1, -1):
        if (i == len(history_list) - 1 and history_list[i] == "CASA END LOGTABLE") or (
            i == 0 and history_list[i] == "CASA START LOGTABLE"
        ):
            history_list.pop(i)
        elif history_list[i].startswith(">"):
            # entry continuation line
            history_list[i - 1] = history_list[i - 1] + history_list[i][1:]
            history_list.pop(i)


def _create_coords(
    helpers: dict, header: fits.header, do_sky_coords: bool
) -> xr.Dataset:
    dir_axes = helpers["dir_axes"]
    dim_map = helpers["dim_map"]
    sphr_dims = (
        [dim_map["l"], dim_map["m"]] if ("l" in dim_map) and ("m" in dim_map) else []
    )
    helpers["sphr_dims"] = sphr_dims
    coords = {}
    coords["time"] = _get_time_values(helpers)
    coords["polarization"] = _get_pol_values(helpers)
    coords["frequency"] = _get_freq_values(helpers)
    coords["velocity"] = (["frequency"], _get_velocity_values(helpers))
    if len(sphr_dims) > 0:
        for i, c in enumerate(["l", "m"]):
            idx = sphr_dims[i]
            cdelt_rad = helpers["cdelt"][idx] * u.Unit(_get_unit(helpers["cunit"][idx]))
            cdelt_rad = abs(cdelt_rad.to("rad").value)
            if c == "l":
                # l values increase to the east
                # l follows RA as far as increasing/decreasing, see AIPS Meme 27,
                # change in alpha definition three lines below Figure 2 and the first
                # of the pair of equations 10.
                cdelt_rad = -cdelt_rad
            helpers[c] = {}
            helpers[c]["cunit"] = "rad"
            helpers[c]["cdelt"] = cdelt_rad
            coords[c] = _compute_linear_world_values(
                naxis=helpers["shape"][idx],
                crpix=helpers["crpix"][idx],
                crval=0.0,
                cdelt=cdelt_rad,
            )
        if do_sky_coords:
            pick = lambda mylist: [mylist[i] for i in sphr_dims]
            my_ret = _compute_world_sph_dims(
                projection=helpers["projection"],
                shape=pick(helpers["shape"]),
                ctype=pick(helpers["ctype"]),
                crpix=pick(helpers["crpix"]),
                crval=pick(helpers["crval"]),
                cdelt=pick(helpers["cdelt"]),
                cunit=pick(helpers["cunit"]),
            )
            for j, i in enumerate(dir_axes):
                helpers["cunit"][i] = my_ret["unit"][j]
                helpers["crval"][i] = my_ret["ref_val"][j]
                helpers["cdelt"][i] = my_ret["inc"][j]
            coords[my_ret["axis_name"][0]] = (["l", "m"], my_ret["value"][0])
            coords[my_ret["axis_name"][1]] = (["l", "m"], my_ret["value"][1])
            helpers["sphr_axis_names"] = tuple(my_ret["axis_name"])
    else:
        # Fourier image
        coords["u"], coords["v"] = _get_uv_values(helpers)
    xds = xr.Dataset(coords=coords)
    return xds


def _get_time_values(helpers):
    return [helpers["obsdate"]["value"]]


def _get_pol_values(helpers):
    # as mapped in casacore Stokes.h
    stokes_map = [
        "Undefined",
        "I",
        "Q",
        "U",
        "V",
        "RR",
        "RL",
        "LR",
        "LL",
        "XX",
        "XY",
        "YX",
        "YY",
        "RX",
        "RY",
        "LX",
        "LY",
        "XR",
        "XL",
        "YR",
        "YL",
        "PP",
        "PQ",
    ]
    idx = helpers["ctype"].index("STOKES")
    if idx >= 0:
        vals = []
        crval = int(helpers["crval"][idx])
        crpix = int(helpers["crpix"][idx])
        cdelt = int(helpers["cdelt"][idx])
        stokes_start_idx = crval - cdelt * crpix
        for i in range(helpers["shape"][idx]):
            stokes_idx = (stokes_start_idx + i) * cdelt
            vals.append(stokes_map[stokes_idx])
        return vals
    else:
        return ["I"]


def _get_freq_values(helpers: dict) -> list:
    vals = []
    ctype = helpers["ctype"]
    if "FREQ" in ctype:
        freq_idx = ctype.index("FREQ")
        helpers["freq_axis"] = freq_idx
        vals = _compute_linear_world_values(
            naxis=helpers["shape"][freq_idx],
            crval=helpers["crval"][freq_idx],
            crpix=helpers["crpix"][freq_idx],
            cdelt=helpers["cdelt"][freq_idx],
        )
        cunit = helpers["cunit"][freq_idx]
        helpers["frequency"] = vals * u.Unit(cunit)
        return vals
    elif "VOPT" in ctype:
        if "restfreq" in helpers:
            restfreq = helpers["restfreq"] * u.Hz
        else:
            raise RuntimeError(
                "Spectral axis in FITS header is velocity, but there is "
                "no rest frequency so converting to frequency is not possible"
            )
        helpers["doppler"] = "Z"
        v_idx = ctype.index("VOPT")
        helpers["freq_idx"] = v_idx
        helpers["freq_axis"] = v_idx
        crval = helpers["crval"][v_idx]
        crpix = helpers["crpix"][v_idx]
        cdelt = helpers["cdelt"][v_idx]
        cunit = helpers["cunit"][v_idx]
        freq, vel = _freq_from_vel(
            crval, cdelt, crpix, cunit, "Z", helpers["shape"][v_idx], restfreq
        )
        helpers["velocity"] = vel["value"] * u.Unit(vel["unit"])
        helpers["crval"][v_idx] = (freq["crval"] * u.Unit(freq["unit"])).to(u.Hz).value
        helpers["cdelt"][v_idx] = (freq["cdelt"] * u.Unit(freq["unit"])).to(u.Hz).value
        return list(freq["value"])
    else:
        return [1420e6]


def _get_velocity_values(helpers: dict) -> list:
    if "velocity" in helpers:
        return helpers["velocity"].to(u.m / u.s).value
    elif "frequency" in helpers:
        v = _compute_velocity_values(
            restfreq=helpers["restfreq"],
            freq_values=helpers["frequency"].to("Hz").value,
            doppler=helpers.get("doppler", "RADIO"),
        )
        helpers["velocity"] = v * (u.m / u.s)
        return v


def _do_multibeam(xds: xr.Dataset, imname: str) -> xr.Dataset:
    """Only run if we are sure there are multiple beams"""
    hdulist = fits.open(imname)
    for hdu in hdulist:
        header = hdu.header
        if "EXTNAME" in header and header["EXTNAME"] == "BEAMS":
            units = (
                u.Unit(header["TUNIT1"]),
                u.Unit(header["TUNIT2"]),
                u.Unit(header["TUNIT3"]),
            )
            nchan = header["NCHAN"]
            npol = header["NPOL"]
            beam_array = np.zeros([1, npol, nchan, 3])
            data = hdu.data
            for t in data:
                beam_array[0, t[4], t[3]] = t[0:3]
            for i in (0, 1, 2):
                beam_array[:, :, :, i] = (
                    (beam_array[:, :, :, i] * units[i]).to("rad").value
                )
            xdb = xr.DataArray(
                beam_array, dims=["time", "polarization", "frequency", "beam_param"]
            )
            xdb = xdb.rename("beam")
            xdb = xdb.assign_coords(beam_param=["major", "minor", "pa"])
            xdb.attrs["units"] = "rad"
            xds["beam"] = xdb
            return xds
    raise RuntimeError(
        "It looks like there should be a BEAMS table but no "
        "such table found in FITS file"
    )


def _get_uv_values(helpers: dict) -> tuple:
    shape = helpers["shape"]
    ctype = helpers["ctype"]
    unit = helpers["cunit"]
    delt = helpers["cdelt"]
    ref_pix = helpers["crpix"]
    ref_val = helpers["crval"]
    for i, axis in enumerate(["UU", "VV"]):
        idx = ctype.index(axis)
        if idx >= 0:
            z = []
            crpix = ref_pix[i]
            crval = ref_val[i]
            cdelt = delt[i]
            for i in range(shape[idx]):
                f = (i - crpix) * cdelt + crval
                z.append(f)
            if axis == "UU":
                u = z
            else:
                v = z
    return u, v


def _add_sky_or_aperture(
    xds: xr.Dataset,
    ary: Union[np.ndarray, da.array],
    dim_order: list,
    helpers: dict,
    has_sph_dims: bool,
) -> xr.Dataset:
    xda = xr.DataArray(ary, dims=dim_order)
    image_type = helpers["btype"]
    unit = helpers["bunit"]
    xda.attrs[_image_type] = image_type
    xda.attrs["units"] = unit
    name = "SKY" if has_sph_dims else "APERTURE"
    xda = xda.rename(name)
    xds[xda.name] = xda
    if helpers["has_mask"]:
        pp = da if type(xda[0].data) == dask.array.core.Array else np
        mask = pp.isnan(xda)
        mask.attrs = {}
        mask = mask.rename("mask0")
        xds["mask0"] = mask
    return xds


def _read_image_array(
    img_full_path: str, chunks: dict, helpers: dict, verbose: bool
) -> da.array:
    # memmap = True allows only part of data to be loaded into memory
    # may also need to pass mode='denywrite'
    # https://stackoverflow.com/questions/35759713/astropy-io-fits-read-row-from-large-fits-file-with-mutliple-hdus
    if isinstance(chunks, dict):
        mychunks = _get_chunk_list(chunks, helpers)
    else:
        raise ValueError(
            f"incorrect type {type(chunks)} for parameter chunks. Must be dict"
        )
    transpose_list, new_axes = _get_transpose_list(helpers)
    data_type = helpers["dtype"]
    rshape = helpers["shape"][::-1]
    full_chunks = mychunks + tuple([1 for rr in range(5) if rr >= len(mychunks)])
    d0slices = []
    blc = tuple(5 * [0])
    trc = tuple(rshape) + tuple([1 for rr in range(5) if rr >= len(mychunks)])
    for d0 in range(blc[0], trc[0], full_chunks[0]):
        d0len = min(full_chunks[0], trc[0] - d0)
        d1slices = []
        for d1 in range(blc[1], trc[1], full_chunks[1]):
            d1len = min(full_chunks[1], trc[1] - d1)
            d2slices = []
            for d2 in range(blc[2], trc[2], full_chunks[2]):
                d2len = min(full_chunks[2], trc[2] - d2)
                d3slices = []
                for d3 in range(blc[3], trc[3], full_chunks[3]):
                    d3len = min(full_chunks[3], trc[3] - d3)
                    d4slices = []
                    for d4 in range(blc[4], trc[4], full_chunks[4]):
                        d4len = min(full_chunks[4], trc[4] - d4)
                        shapes = tuple(
                            [d0len, d1len, d2len, d3len, d4len][: len(rshape)]
                        )
                        starts = tuple([d0, d1, d2, d3, d4][: len(rshape)])
                        delayed_array = dask.delayed(_read_image_chunk)(
                            img_full_path, shapes, starts
                        )
                        d4slices += [da.from_delayed(delayed_array, shapes, data_type)]
                    d3slices += (
                        [da.concatenate(d4slices, axis=4)]
                        if len(rshape) > 4
                        else d4slices
                    )
                d2slices += (
                    [da.concatenate(d3slices, axis=3)] if len(rshape) > 3 else d3slices
                )
            d1slices += (
                [da.concatenate(d2slices, axis=2)] if len(rshape) > 2 else d2slices
            )
        d0slices += [da.concatenate(d1slices, axis=1)] if len(rshape) > 1 else d1slices
    ary = da.concatenate(d0slices, axis=0)
    ary = da.expand_dims(ary, new_axes)
    return ary.transpose(transpose_list)


def _get_chunk_list(chunks: dict, helpers: dict) -> tuple:
    ret_list = list(helpers["shape"])[::-1]
    axis = 0
    ctype = helpers["ctype"]
    for c in ctype[::-1]:
        if c.startswith("RA"):
            if "l" in chunks:
                ret_list[axis] = chunks["l"]
        elif c.startswith("DEC"):
            if "m" in chunks:
                ret_list[axis] = chunks["m"]
        elif c.startswith("FREQ") or c.startswith("VOPT") or c.startswith("VRAD"):
            if "frequency" in chunks:
                ret_list[axis] = chunks["frequency"]
        elif c.startswith("STOKES"):
            if "polarization" in chunks:
                ret_list[axis] = chunks["polarization"]
        else:
            raise RuntimeError(f"Unhandled coordinate type {c}")
        axis += 1
    return tuple(ret_list)


def _get_transpose_list(helpers: dict) -> tuple:
    ctype = helpers["ctype"]
    transpose_list = 5 * [-1]
    # time axis
    transpose_list[0] = 4
    new_axes = [4]
    last_axis = 3
    not_covered = ["l", "m", "u", "v", "s", "f"]
    for i, c in enumerate(ctype[::-1]):
        b = c.lower()
        if b.startswith("ra") or b.startswith("uu"):
            transpose_list[3] = i
            not_covered.remove("l")
            not_covered.remove("u")
        elif b.startswith("dec") or b.startswith("vv"):
            transpose_list[4] = i
            not_covered.remove("m")
            not_covered.remove("v")
        elif (
            b.startswith("frequency")
            or b.startswith("freq")
            or b.startswith("vopt")
            or b.startswith("vrad")
        ):
            transpose_list[2] = i
            not_covered.remove("f")
        elif b.startswith("stok"):
            transpose_list[1] = i
            not_covered.remove("s")
        else:
            raise RuntimeError(f"Unhandled axis name {c}")
    h = {"l": 3, "m": 4, "u": 3, "v": 4, "f": 2, "s": 1}
    for p in not_covered:
        transpose_list[h[p]] = last_axis
        new_axes.append(last_axis)
        last_axis -= 1
    new_axes.sort()
    if transpose_list.count(-1) > 0:
        raise RuntimeError(f"Logic error: axes {axes}, transpose_list {transpose_list}")
    return transpose_list, new_axes


def _read_image_chunk(img_full_path, shapes: tuple, starts: tuple) -> np.ndarray:
    hdulist = fits.open(img_full_path, memmap=True)
    s = []
    for start, length in zip(starts, shapes):
        s.append(slice(start, start + length))
    t = tuple(s)
    z = hdulist[0].data[t]
    hdulist.close()
    # delete to avoid having a reference to a mem-mapped hdulist
    del hdulist
    return z



import os
from typing import Dict, Union

import numpy as np
import pandas as pd
import xarray as xr

from ._tables.read import load_generic_table, read_flat_col_chunk
from ._tables.table_query import open_query, open_table_ro
from xradio._utils.list_and_array import unique_1d


def describe_ms(
    infile: str, mode: str = "summary", rowmap: Union[dict, None] = None
) -> Union[pd.DataFrame, Dict]:
    """
    Summarize the contents of an MS directory in casacore table format

    Parameters
    ----------
    infile : str
        input MS filename
    mode : str (Default value = "summary")
        type of information returned ('summary', 'flat', 'expanded').
        'summary' returns a pandas dataframe that is nice for displaying in notebooks
        etc. 'flat' returns a list of tuples of (ddi, row, chan, pol). 'expanded'
        returns a list of tuples of (ddi, time, baseline, chan, pol). These latter two
        are good for trying to determine chunk size for read_ms(expand=True/False). (Default value = "summary")
    rowmap : Union[dict, None] (Default value = None)
        dict of DDI to tuple of (row indices, channel indices). Returned
        by ms_selection function. Default None ignores selections

    Returns
    -------
    Union[pd.DataFrame, Dict]
        summary as a pd dataframe
    """
    infile = os.path.expanduser(infile)  # does nothing if $HOME is unknown
    if not os.path.isdir(infile):
        raise ValueError(f"invalid input filename to describe_ms: {infile}")
    if mode not in [
        "summary",
        "flat",
        "expanded",
    ]:
        raise ValueError("invalid mode, must be summary, flat or expanded")

    ddi_xds = load_generic_table(infile, "DATA_DESCRIPTION")
    ddis = list(ddi_xds.row.values) if rowmap is None else list(rowmap.keys())
    summary: Union[pd.DataFrame, Dict] = []
    if mode == "summary":
        summary = pd.DataFrame([])

    all_sdf = []
    with open_table_ro(infile) as tb_tool:
        for ddi in ddis:
            taql = f"select * from $tb_tool where DATA_DESC_ID = {ddi}"
            with open_query(tb_tool, taql) as query_per_ddi:
                sdf = populate_ms_descr(
                    infile, mode, query_per_ddi, summary, ddi, ddi_xds
                )
                all_sdf.append(sdf)

    if mode == "summary":
        summary = pd.DataFrame(all_sdf)
        summary = summary.set_index("ddi").sort_index()
    else:
        summary = dict(summary)
    return summary


def populate_ms_descr(
    infile: str,
    mode: str,
    query_per_ddi,
    summary: dict,
    ddi: int,
    ddi_xds: xr.Dataset,
    rowmap: Union[Dict, None] = None,
) -> pd.DataFrame:
    """
    Adds information from the time and baseline (antenna1+antenna2)
    columns as well as channel and polarizations, based on a taql
    query.

    Parameters
    ----------
    infile : str
        input table/MS path
    mode : str
        mode (as in describe_ms())
    query_per_ddi :
        a TaQL query with data per individual DDI
    summary : Dict
        summary dict being populated
    ddi_xds : xr.Dataset
        final summary object being populated from the invividual sdf's

    Returns
    -------
    pd.DataFrame
    """
    spw_ids = ddi_xds.SPECTRAL_WINDOW_ID.values
    pol_ids = ddi_xds.POLARIZATION_ID.values
    sdf = {
        "ddi": ddi,
        "spw_id": spw_ids[ddi],
        "pol_id": pol_ids[ddi],
        "rows": query_per_ddi.nrows(),
    }

    # figure out characteristics of main table from select subtables (must all be present)
    spw_xds = load_generic_table(infile, "SPECTRAL_WINDOW")
    pol_xds = load_generic_table(infile, "POLARIZATION")

    if mode in ["expanded", "summary"]:
        times = (
            query_per_ddi.getcol("TIME")
            if rowmap is None
            else read_flat_col_chunk(infile, "TIME", (1,), rowmap[ddi][0], 0, 0)
        )
        baselines = [
            (
                query_per_ddi.getcol(rr)[:, None]
                if rowmap is None
                else read_flat_col_chunk(infile, rr, (1,), rowmap[ddi][0], 0, 0)
            )
            for rr in ["ANTENNA1", "ANTENNA2"]
        ]
        sdf.update(
            {
                "times": len(unique_1d(times)),
                "baselines": len(np.unique(np.hstack(baselines), axis=0)),
            }
        )

    chans = spw_xds.NUM_CHAN.values
    pols = pol_xds.NUM_CORR.values
    sdf.update(
        {
            "chans": (
                chans[spw_ids[ddi]]
                if (rowmap is None) or (rowmap[ddi][1] is None)
                else len(rowmap[ddi][1])
            ),
            "pols": pols[pol_ids[ddi]],
        }
    )
    sdf["size_MB"] = np.ceil(
        (sdf["rows"] * sdf["chans"] * sdf["pols"] * 10) / 1024**2
    ).astype(int)

    if rowmap is not None:
        sdf["rows"] = len(rowmap[ddi][0])

    if mode == "summary":
        summary = pd.concat(
            [summary, pd.DataFrame(sdf, index=[str(ddi)])], axis=0, sort=False
        )
    elif mode == "flat":
        summary += [(ddi, (sdf["rows"], sdf["chans"], sdf["pols"]))]
    else:
        summary += [((ddi, sdf["times"], sdf["baselines"], sdf["chans"], sdf["pols"]))]

    return sdf
